/*
 *   -- micMAGMA (version 0.1) --
 *      Univ. of Tennessee, Knoxville
 *      Univ. of California, Berkeley
 *      Univ. of Colorado, Denver
 *      April 2012
 *
 * @generated d Thu May  2 22:50:36 2013
 */

#include <mkl.h>

#include <mkl_trans.h>
#include <scif.h>
#include <fcntl.h>
#include <offload.h>

//#ifdef __MIC__
//#define _M512_TYPED
//#include <lmmintrin.h>
//#endif
#include <immintrin.h>


#include "magma.h"
//#include "magma_d_bindings.h"
#include "magma_d_server.h"



void call_dgemm(magma_mic_dgemm_param dgemm_param)
{
    magma_trans_t transa, transb;
    magma_int_t m, n, k, lda, ldb, ldc;
    double alpha, beta, *a, *b, *c;
    
    transa = dgemm_param.transa;
    transb = dgemm_param.transb;
    m = dgemm_param.m;
    n = dgemm_param.n;
    k = dgemm_param.k;
    alpha = dgemm_param.alpha;
    a = (double *)dgemm_param.a;
    lda = dgemm_param.lda;
    b = (double *)dgemm_param.b;
    ldb = dgemm_param.ldb;
    beta = dgemm_param.beta;
    c = (double *)dgemm_param.c;
    ldc = dgemm_param.ldc;

    //printf("DGEMM: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);

    dgemm( lapack_const(transa), lapack_const(transb), 
           &m, &n, &k, (const double *)&alpha, (double *)a, &lda, 
           (double *)b, &ldb, (const double *)&beta, 
           (double *)c, &ldc);
}

void call_dsymm(magma_mic_dsymm_param dsymm_param)
{
    magma_side_t side;
    magma_uplo_t uplo;
    magma_int_t m, n, lda, ldb, ldc;
    double alpha, beta, *a, *b, *c;

    side = dsymm_param.side;
    uplo = dsymm_param.uplo;
    m = dsymm_param.m;
    n = dsymm_param.n;
    alpha = dsymm_param.alpha;
    a = (double *)dsymm_param.a;
    lda = dsymm_param.lda;
    b = (double *)dsymm_param.b;
    ldb = dsymm_param.ldb;
    beta = dsymm_param.beta;
    c = (double *)dsymm_param.c;
    ldc = dsymm_param.ldc;

    dsymm( lapack_const(side), lapack_const(uplo),
           &m, &n, (const double *)&alpha, (double *)a, &lda,
           (double *)b, &ldb, (const double *)&beta,
           (double *)c, &ldc);
}

void call_dgemv(magma_mic_dgemv_param dgemv_param)
{
    magma_trans_t transa;
    magma_int_t m, n, lda, incx, incy;
    double alpha, beta, *a, *x, *y;
    
    transa = dgemv_param.transa;
    m      = dgemv_param.m;
    n      = dgemv_param.n;
    alpha  = dgemv_param.alpha;
    a      = (double *)dgemv_param.a;
    lda    = dgemv_param.lda;
    x      = (double *)dgemv_param.x;
    incx   = dgemv_param.incx;
    beta   = dgemv_param.beta;
    y      = (double *)dgemv_param.y;
    incy   = dgemv_param.incy;
    
    //printf("DGEMV: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);
    
    dgemv( lapack_const(transa), &m, &n, (const double *)&alpha,
           (double *)a, &lda, (double *)x, &incx,
           (const double *)&beta, (double *)y, &incy);
}

void call_dsymv(magma_mic_dsymv_param symv_param)
{
    magma_uplo_t uplo;
    magma_int_t n, lda, incx, incy;
    double alpha, beta, *a, *x, *y;
    
    uplo   = symv_param.uplo;
    n      = symv_param.n;
    alpha  = symv_param.alpha;
    a      = (double *)symv_param.a;
    lda    = symv_param.lda;
    x      = (double *)symv_param.x;
    incx   = symv_param.incx;
    beta   = symv_param.beta;
    y      = (double *)symv_param.y;
    incy   = symv_param.incy;
    
    dsymv( lapack_const(uplo), &n, (const double *)&alpha,
           (double *)a, &lda, (double *)x, &incx,
           (const double *)&beta, (double *)y, &incy);
}

void call_dsymv2(magma_mic_dsymv2_param symv_param)
{
    magma_uplo_t uplo;
    magma_int_t n, lda, incx, incy;
    double alpha, beta, *a, *x, *y;
    
    uplo   = symv_param.uplo;
    n      = symv_param.n;
    alpha  = symv_param.alpha;
    a      = (double *)symv_param.a;
    lda    = symv_param.lda;
    x      = (double *)symv_param.x;
    incx   = symv_param.incx;
    beta   = symv_param.beta;
    y      = (double *)symv_param.y;
    incy   = symv_param.incy;
    
    dsymv( lapack_const(uplo), &n, (const double *)&alpha,
           (double *)a, &lda, (double *)x, &incx,
           (const double *)&beta, (double *)y, &incy);
}

void call_dtrsm(magma_mic_dtrsm_param dtrsm_param)
{
    magma_side_t side;
    magma_uplo_t uplo;
    magma_trans_t transa;
    magma_diag_t diag;
    magma_int_t m, n, lda, ldb;
    double alpha, *a, *b;
    
    side = dtrsm_param.side;
    uplo = dtrsm_param.uplo;
    transa = dtrsm_param.transa;
    diag = dtrsm_param.diag;
    m = dtrsm_param.m;
    n = dtrsm_param.n;
    alpha = dtrsm_param.alpha;
    a = (double *)dtrsm_param.a;
    lda = dtrsm_param.lda;
    b = (double *)dtrsm_param.b;
    ldb = dtrsm_param.ldb;
    
    //printf("DTRSM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
    dtrsm( lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
           &m, &n, (const double *)&alpha, (double *)a, &lda, 
           (double *)b, &ldb);
}

void call_dtrsv(magma_mic_dtrsv_param dtrsv_param)
{
    magma_uplo_t uplo;
    magma_trans_t trans;
    magma_diag_t diag;
    magma_int_t n, lda, incx;
    double alpha, *a, *dx;
    
    uplo  = dtrsv_param.uplo;
    trans = dtrsv_param.trans;
    diag  = dtrsv_param.diag;
    n     = dtrsv_param.n;
    a     = (double *)dtrsv_param.a;
    lda   = dtrsv_param.lda;
    dx    = (double *)dtrsv_param.dx;
    incx  = dtrsv_param.incx;
    
    dtrsv( lapack_const(uplo), lapack_const(trans), lapack_const(diag),
           &n, (double *)a, &lda, (double *)dx, &incx);
}

void call_dtrmv(magma_mic_dtrmv_param trmv_param)
{
    magma_uplo_t uplo;
    magma_trans_t trans;
    magma_diag_t diag;
    magma_int_t n, lda, incx;
    double alpha, *a, *dx;
    
    uplo  = trmv_param.uplo;
    trans = trmv_param.trans;
    diag  = trmv_param.diag;
    n     = trmv_param.n;
    a     = (double *)trmv_param.a;
    lda   = trmv_param.lda;
    dx    = (double *)trmv_param.dx;
    incx  = trmv_param.incx;
    
    dtrmv( lapack_const(uplo), lapack_const(trans), lapack_const(diag),
           &n, (double *)a, &lda, (double *)dx, &incx);
}


void call_dtrmm(magma_mic_dtrmm_param dtrmm_param)
{
    magma_side_t side;
    magma_uplo_t uplo;
    magma_trans_t transa;
    magma_diag_t diag;
    magma_int_t m, n, lda, ldb;
    double alpha, *a, *b;

    side = dtrmm_param.side;
    uplo = dtrmm_param.uplo;
    transa = dtrmm_param.transa;
    diag = dtrmm_param.diag;
    m = dtrmm_param.m;
    n = dtrmm_param.n;
    alpha = dtrmm_param.alpha;
    a = (double *)dtrmm_param.a;
    lda = dtrmm_param.lda;
    b = (double *)dtrmm_param.b;
    ldb = dtrmm_param.ldb;

    //printf("DTRMM: %c %c %c %c %d %d %f 0x%lx %d 0x%lx %d\n", side, uplo, transa, diag, m, n, alpha, (off_t)a, lda, (off_t)b, ldb);
    dtrmm( lapack_const(side), lapack_const(uplo), lapack_const(transa), lapack_const(diag),
           &m, &n, (const double *)&alpha, (double *)a, &lda, 
           (double *)b, &ldb);
}


//magmablas our dsyrk
void magmablas_dsyrk_mic(
    char uplo, char trans, magma_int_t n, magma_int_t k, magma_int_t nb,
    double alpha, double *a, magma_int_t lda, magma_int_t aoff,
    double beta,  double *c, magma_int_t ldc, magma_int_t offset)
{
    #define min(a, b) ((a) < (b) ? (a) : (b))
    #define dA(i, j) (a + (i) + (j)*lda + (aoff) )
    #define dC(i, j) (c + (i) + (j)*ldc)

    double cbeta  = MAGMA_D_MAKE( beta, 0. );
    double calpha = MAGMA_D_MAKE( alpha, 0. );

    magma_int_t m, ib, ioff;
    magma_int_t blockoffset = offset % nb;

    // loop over all blocks and does A*A'
    // blockoffset is offset within first block; for subsequent blocks it is 0
    for( magma_int_t i = 0; i < n; i += ib ) {
        ib     = min( nb-blockoffset, n-i );  // block size
        ioff   = i + offset;                  // global index in parent matrix
        // C[i:n,i] += A[i:n,0] * A[i,0]'
        m = n - i;
        dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), &m, &ib, &k,
               (const double *)&calpha, (double *)(dA(i,0)),       &lda,
               (double *)(dA(i,0)),       &lda,
               (const double *)&cbeta, (double *)(dC(ioff,ioff)), &ldc );
        blockoffset = 0;
    }
}


void call_dsyrk(magma_mic_dsyrk_param dsyrk_param)
{
    magma_uplo_t uplo;
    magma_trans_t trans;
    magma_int_t n, k, lda, ldc;
    double alpha, beta;
    double *a, *c;

    uplo = dsyrk_param.uplo;
    trans = dsyrk_param.trans;
    n = dsyrk_param.n;
    k = dsyrk_param.k;
    alpha = dsyrk_param.alpha;
    a = (double *)dsyrk_param.a;
    lda = dsyrk_param.lda;
    beta = dsyrk_param.beta;
    c = (double *)dsyrk_param.c;
    ldc = dsyrk_param.ldc;

    //printf("DSYRK: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo, trans, n, k, alpha, (off_t)a, lda, beta, (off_t)c, ldc);

    //dsyrk( lapack_const(uplo), lapack_const(trans), 
    //       &n, &k, &alpha, (double *)a, &lda, 
    //       &beta, (double *)c, &ldc);
    
    magma_int_t nb = 960;
    magmablas_dsyrk_mic('L', 'N', 
                     n, k, nb, alpha, a, lda, 0, 
                     beta, c, ldc, 0);
}

void magmablas_dsyrk_mmic(
    char uplo, char trans, magma_int_t n, magma_int_t k,
    double alpha, double *a, magma_int_t lda, magma_int_t aoff,
    double beta,  double *c, magma_int_t ldc, magma_int_t offset,
    magma_int_t nmic, magma_int_t nb, magma_int_t my_id)
{
    #define dA(i, j) (a + (i) + (j)*lda + (aoff) )
    #define dC(i, j) (c + (i) + (j)*ldc)

    double cbeta  = MAGMA_D_MAKE( beta,  0. );
    double calpha = MAGMA_D_MAKE( alpha, 0. );

    magma_int_t rows, ib, ioff, iblock, idev, di, s;

    // loop over all blocks
    // Faster to have two loops: first does A*B', second does B*A'
    // blockoffset is offset within first block; for subsequent blocks it is 0
    magma_int_t blockoffset = offset % nb;
    for( magma_int_t i = 0; i < n; i += ib ) {
        ib     = min( nb-blockoffset, n-i );  // block size
        ioff   = i + offset;                  // global index in parent matrix
        iblock = (ioff / nb) / nmic;          // local block id
        idev   = (ioff / nb) % nmic;          // device with this block
        di     = iblock*nb + blockoffset;     // local index in parent matrix

        // C[i:n,i] += A[i:n,0] * B[i,0]'
        //printf( "dgemm  n=%4d, ib=%4d, k=%4d, i=%4d\n", n-i, ib, k, i );
        if(idev == my_id) {
            rows = n - i;
            dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), &rows, &ib, &k,
                    (const double *)&calpha, (double *)dA(i,0), &lda,
                    (double *)dA(i,0), &lda,
                    (const double *)&cbeta, (double *)dC(ioff,di), &ldc);
        }

        blockoffset = 0;
    }
}

void call_dsyrk_mmic(magma_mic_dsyrk_mmic_param dsyrk_mmic_param)
{
    magma_uplo_t uplo;
    magma_trans_t trans;
    magma_int_t n, k, lda, ldb, ldc, aoff, boff, offset, nb, my_id, nmic;
    double alpha, beta;
    double *a, *b, *c;

    uplo = dsyrk_mmic_param.uplo;
    trans = dsyrk_mmic_param.trans;
    n = dsyrk_mmic_param.n;
    k = dsyrk_mmic_param.k;
    alpha = dsyrk_mmic_param.alpha;
    a = (double *)dsyrk_mmic_param.a;
    lda = dsyrk_mmic_param.lda;
    aoff = dsyrk_mmic_param.aoff;
    beta = dsyrk_mmic_param.beta;
    c = (double *)dsyrk_mmic_param.c;
    ldc = dsyrk_mmic_param.ldc;
    offset = dsyrk_mmic_param.offset;
    nmic = dsyrk_mmic_param.nmic;
    nb = dsyrk_mmic_param.nb;
    my_id = dsyrk_mmic_param.my_id;

    //printf("DSYRK_MMIC: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo, trans, n, k, alpha, (off_t)a, lda, beta, (off_t)c, ldc);

    magmablas_dsyrk_mmic(
        uplo, trans, n, k,
        alpha, a, lda, aoff,
        beta, c, ldc, offset,
        nmic, nb, my_id);
}



void magmablas_dsyr2k_mmic(
    char uplo, char trans, magma_int_t n, magma_int_t k,
    double alpha, double *a, magma_int_t lda, magma_int_t aoff,
    double *b, magma_int_t ldb, magma_int_t boff,
    double beta,           double *c, magma_int_t ldc,  magma_int_t offset,
    magma_int_t nmic, magma_int_t nb, magma_int_t my_id)
{
    #define dA(i, j) (a + (i) + (j)*lda + (aoff) )
    #define dB(i, j) (b + (i) + (j)*ldb + (boff) )
    #define dC(i, j) (c + (i) + (j)*ldc)

    double cbeta = MAGMA_D_MAKE( beta, 0. );

    magma_int_t rows, ib, ioff, iblock, idev, di, s;

    // loop over all blocks
    // Faster to have two loops: first does A*B', second does B*A'
    // blockoffset is offset within first block; for subsequent blocks it is 0
    magma_int_t blockoffset = offset % nb;
    for( magma_int_t i = 0; i < n; i += ib ) {
        ib     = min( nb-blockoffset, n-i );  // block size
        ioff   = i + offset;                  // global index in parent matrix
        iblock = (ioff / nb) / nmic;          // local block id
        idev   = (ioff / nb) % nmic;          // device with this block
        di     = iblock*nb + blockoffset;     // local index in parent matrix

        // C[i:n,i] += A[i:n,0] * B[i,0]'
        //printf( "dgemm  n=%4d, ib=%4d, k=%4d, i=%4d\n", n-i, ib, k, i );
        if(idev == my_id) {
            rows = n - i;
            dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), &rows, &ib, &k,
                    (const double *)&alpha, (double *)dA(i,0), &lda,
                    (double *)dB(i,0), &ldb,
                    (const double *)&cbeta, (double *)dC(ioff,di), &ldc);
        }

        blockoffset = 0;
    }
    blockoffset = offset % nb;
    for( magma_int_t i = 0; i < n; i += ib ) {
        ib     = min( nb-blockoffset, n-i );  // block size
        ioff   = i + offset;                  // global index in parent matrix
        iblock = (ioff / nb) / nmic;          // local block id
        idev   = (ioff / nb) % nmic;          // device with this block
        di     = iblock*nb + blockoffset;     // local index in parent matrix

        // C[i:n,i] += B[i:n,0] * A[i,0]'
        //printf( "dgemm  n=%4d, ib=%4d, k=%4d, i=%4d\n", n-i, ib, k, i );
        if(idev == my_id) {
            rows = n - i;
            dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans), &rows, &ib, &k,
                    (const double *)&alpha, (double *)dB(i,0), &ldb,
                    (double *)dA(i,0), &lda,
                    (const double *)&cbeta, (double *)dC(ioff,di), &ldc);
        }
        blockoffset = 0;
    }
}


void call_dsyr2k_mmic(magma_mic_dsyr2k_mmic_param dsyr2k_mmic_param)
{
    magma_uplo_t uplo;
    magma_trans_t trans;
    magma_int_t n, k, lda, ldb, ldc, aoff, boff, offset, nb, my_id, nmic;
    double alpha;
    double beta;
    double *a, *b, *c;

    uplo = dsyr2k_mmic_param.uplo;
    trans = dsyr2k_mmic_param.trans;
    n = dsyr2k_mmic_param.n;
    k = dsyr2k_mmic_param.k;
    alpha = dsyr2k_mmic_param.alpha;
    a = (double *)dsyr2k_mmic_param.a;
    lda = dsyr2k_mmic_param.lda;
    aoff = dsyr2k_mmic_param.aoff;
    b = (double *)dsyr2k_mmic_param.b;
    ldb = dsyr2k_mmic_param.ldb;
    boff = dsyr2k_mmic_param.boff;
    beta = dsyr2k_mmic_param.beta;
    c = (double *)dsyr2k_mmic_param.c;
    ldc = dsyr2k_mmic_param.ldc;
    offset = dsyr2k_mmic_param.offset;
    nmic = dsyr2k_mmic_param.nmic;
    nb = dsyr2k_mmic_param.nb;
    my_id = dsyr2k_mmic_param.my_id;

    //printf("DSYR2K_MMIC: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo, trans, n, k, alpha, (off_t)a, lda, beta, (off_t)c, ldc);

    magmablas_dsyr2k_mmic(
        uplo, trans, n, k,
        alpha, a, lda, aoff,
        b, ldb, boff,
        beta, c, ldc, offset,
        nmic, nb, my_id);
}

void call_dsyr2k(magma_mic_dsyr2k_param dsyr2k_param)
{
    magma_uplo_t uplo;
    magma_trans_t trans;
    magma_int_t n, k, lda, ldb, ldc;
    double alpha;
    double beta;
    double *a, *b, *c;

    uplo = dsyr2k_param.uplo;
    trans = dsyr2k_param.trans;
    n = dsyr2k_param.n;
    k = dsyr2k_param.k;
    alpha = dsyr2k_param.alpha;

    a = (double *)dsyr2k_param.a;
    lda = dsyr2k_param.lda;

    b = (double *)dsyr2k_param.b;
    ldb = dsyr2k_param.ldb;

    beta = dsyr2k_param.beta;
    c = (double *)dsyr2k_param.c;
    ldc = dsyr2k_param.ldc;

    //printf("DSYR2K: %c %c %d %d %f 0x%lx %d %f 0x%lx %d\n", uplo,trans, n,k, alpha, (off_t)a,lda, beta, (off_t)c, ldc);
    
    dsyr2k( lapack_const(uplo), lapack_const(trans),
            &n, &k, (double *)&alpha, (double *)a, &lda,
            (double *)b, &ldb,
            &beta, (double *)c, &ldc);
}


void call_dlarft(magma_mic_dlarft_param dlarft_param)
{
    int direct, storev;
    magma_int_t n, k, ldv, ldt; 
    double *v, *tau, *t;
    
    direct = dlarft_param.direct;
    storev = dlarft_param.storev;
    n = dlarft_param.n;
    k = dlarft_param.k;
    v = (double *)dlarft_param.v;
    ldv = dlarft_param.ldv;
    tau = (double *)dlarft_param.tau;
    t = (double *)dlarft_param.t;
    ldt = dlarft_param.ldt;
    
    //printf("DLARFT: %c %c %d %d 0x%lx %d ox%lx 0x%lx %d\n", direct, storev, n, k, (off_t)v, ldv, (off_t)tau, (off_t)t, ldt);
    
    dlarft( lapack_const(direct), lapack_const(storev),
            &n, &k, (double *)v, &ldv, 
            (double *)tau, (double *)t, &ldt);
}


void call_dlarfb(magma_mic_dlarfb_param dlarfb_param)
{
    magma_int_t side, trans, direct, storev;
    magma_int_t m, n, k, ldv, ldt, ldc, ldwork;
    double *v, *t, *c, *work;

    side = dlarfb_param.side;
    trans = dlarfb_param.trans;
    direct = dlarfb_param.direct;
    storev = dlarfb_param.storev;
    m = dlarfb_param.m;
    n = dlarfb_param.n;
    k = dlarfb_param.k;
    v = (double *)dlarfb_param.v;
    ldv = dlarfb_param.ldv;
    t = (double *)dlarfb_param.t;
    ldt = dlarfb_param.ldt;
    c = (double *)dlarfb_param.c;
    ldc = dlarfb_param.ldc;
    work = (double *)dlarfb_param.work;
    ldwork = dlarfb_param.ldwork;

    //printf("DLARFB: %c %c %c %c %d %d %d 0x%lx %d ox%lx %d 0x%lx %d 0x%lx %d\n", side, trans, direct, storev, m, n, k, (off_t)v, ldv, (off_t)t, ldt, (off_t)c, ldc, (off_t)work, ldwork); 
    dlarfb(lapack_const(side), lapack_const(trans), lapack_const(direct), lapack_const(storev),
                    &m, &n, &k, (double *)v, &ldv, (double *)t, &ldt, 
                    (double *)c, &ldc, (double *)work, &ldwork);
    return;

    //magmablas
    double c_zero    = MAGMA_D_ZERO;
    double c_one     = MAGMA_D_ONE;
    double c_neg_one = MAGMA_D_NEG_ONE;

    /* Function Body */
    if (m <= 0 || n <= 0) {
        return ;//MAGMA_SUCCESS;
    }

    // opposite of trans
    //char transt;
    magma_int_t transt;
    if (trans == MagmaNoTrans)
        transt = MagmaTrans;
    else
        transt = MagmaNoTrans;

    // whether T is upper or lower triangular
    //char uplo;
    magma_int_t uplo;
    if (direct == MagmaForward)
        uplo = MagmaUpper;
    else
        uplo = MagmaLower;

    // whether V is stored transposed or not
    //    char notransV, transV;
    magma_int_t notransV, transV;
    if (storev == MagmaColumnwise) {
        notransV = MagmaNoTrans;
        transV   = MagmaTrans;
    }
    else {
        notransV = MagmaTrans;
        transV   = MagmaNoTrans;
    }

    if ( side  == MagmaLeft ) {
        // Form H C or H^H C
        // Comments assume H C. When forming H^H C, T gets transposed via transt.

        // W = C^H V
        dgemm( lapack_const(MagmaTrans), lapack_const(notransV),
                        &n, &k, &m,
                        (const double *)&c_one,  (double *)c,    &ldc,
                        (double *)v,    &ldv,
                        (const double *)&c_zero, (double *)work, &ldwork);

        // W = W T^H = C^H V T^H
        dtrmm( lapack_const(MagmaRight), lapack_const(uplo), lapack_const(transt), lapack_const(MagmaNonUnit),
                        &n, &k,
                        (const double *)&c_one, (double *)t,    &ldt,
                        (double *)work, &ldwork);

        // C = C - V W^H = C - V T V^H C = (I - V T V^H) C = H C
        dgemm( lapack_const(notransV), lapack_const(MagmaTrans),
                        &m, &n, &k,
                        (const double *)&c_neg_one, (double *)v, &ldv,
                        (double *)work, &ldwork,
                        (const double *)&c_one,     (double *)c,    &ldc);
    }
    else {
        // Form C H or C H^H
        // Comments assume C H. When forming C H^H, T gets transposed via trans.

        // W = C V
        dgemm( lapack_const(MagmaNoTrans), lapack_const(notransV),
                        &m, &k, &n,
                        (const double *)&c_one,  (double *)c,    &ldc,
                        (double *)v,    &ldv,
                        (const double *)&c_zero, (double *)work, &ldwork);

        // W = W T = C V T
        dtrmm( lapack_const(MagmaRight), lapack_const(uplo), lapack_const(trans), lapack_const(MagmaNonUnit),
                        &m, &k,
                        (const double *)&c_one, (double *)t,    &ldt,
                        (double *)work, &ldwork);

        // C = C - W V^H = C - C V T V^H = C (I - V T V^H) = C H
        dgemm( lapack_const(MagmaNoTrans), lapack_const(transV),
                        &m, &n, &k,
                        (const double *)&c_neg_one, (double *)work, &ldwork,
                        (double *)v,    &ldv,
                        (const double *)&c_one, (double *)c, &ldc);
    }
}

void call_dlarfb_gemm(magma_mic_dlarfb_gemm_param dlarfb_gemm_param)
{
    magma_int_t side, trans, direct, storev;
    magma_int_t m, n, k, ldv, ldt, ldc, ldwork, ldworkvt;
    double *v, *t, *c, *work, *workvt;
    
    side = dlarfb_gemm_param.side;
    trans = dlarfb_gemm_param.trans;
    direct = dlarfb_gemm_param.direct;
    storev = dlarfb_gemm_param.storev;
    m = dlarfb_gemm_param.m;
    n = dlarfb_gemm_param.n;
    k = dlarfb_gemm_param.k;
    v = (double *)dlarfb_gemm_param.v;
    ldv = dlarfb_gemm_param.ldv;
    t = (double *)dlarfb_gemm_param.t;
    ldt = dlarfb_gemm_param.ldt;
    c = (double *)dlarfb_gemm_param.c;
    ldc = dlarfb_gemm_param.ldc;
    work = (double *)dlarfb_gemm_param.work;
    ldwork = dlarfb_gemm_param.ldwork;
    workvt = (double *)dlarfb_gemm_param.workvt;
    ldworkvt = dlarfb_gemm_param.ldworkvt;
    
    //printf("DLARFB: %c %c %c %c %d %d %d 0x%lx %d ox%lx %d 0x%lx %d 0x%lx %d\n", side, trans, direct, storev, m, n, k, (off_t)v, ldv, (off_t)t, ldt, (off_t)c, ldc, (off_t)work, ldwork); 
    //dlarfb( lapack_const(side), lapack_const(trans), lapack_const(direct), lapack_const(storev),
    //        &m, &n, &k, (double *)v, &ldv, (double *)t, &ldt, 
    //        (double *)c, &ldc, (double *)work, &ldwork);
    //return;
        
    double c_zero    = MAGMA_D_ZERO;
    double c_one     = MAGMA_D_ONE;
    double c_neg_one = MAGMA_D_NEG_ONE;

    /* Function Body */
    if (m <= 0 || n <= 0) {
        return;
    }

    // opposite of trans
    magma_int_t transt;
    if (trans == MagmaNoTrans)
        transt = MagmaTrans;
    else
        transt = MagmaNoTrans;
    
    // whether T is upper or lower triangular
    magma_int_t uplo;
    if (direct == MagmaForward)
        uplo = MagmaUpper;
    else
        uplo = MagmaLower;
    
    // whether V is stored transposed or not
    magma_int_t notransV, transV;
    if (storev == MagmaColumnwise) {
        notransV = MagmaNoTrans;
        transV   = MagmaTrans;
    }
    else {
        notransV = MagmaTrans;
        transV   = MagmaNoTrans;
    }

    if ( side  == MagmaLeft) {
        // Form H C or H' C
        // Comments assume H C. 
        // When forming H' C, T gets transposed via transt for m>=n or by trans for m<n.
        
        // W = C' V
        dgemm( lapack_const(MagmaTrans), lapack_const(notransV),
                     &n, &k, &m,
                     (const double *)&c_one,  (double *)c, &ldc,
                             (double *)v, &ldv,
                     (const double *)&c_zero, (double *)work, &ldwork);

        if(m<n){
            // W2 = V T 
            dgemm( lapack_const(notransV), lapack_const(trans),
                         &m, &k, &k,
                         (const double *)&c_one,  (double *)v, &ldv,
                                 (double *)t, &ldt,
                         (const double *)&c_zero, (double *)workvt, &ldworkvt);
                        
            // C = C - W2 W' = C - V T V' C = (I - V T V') C = H C 
            dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaTrans),
                         &m, &n, &k,
                         (const double *)&c_neg_one, (double *)workvt, &ldworkvt,
                                    (double *)work, &ldwork,
                         (const double *)&c_one, (double *)c, &ldc);
        }else{
            // W2 = W T' = C' V T'
            dgemm( lapack_const(MagmaNoTrans), lapack_const(transt),
                         &n, &k, &k,
                         (const double *)&c_one,  (double *)work, &ldwork,
                                 (double *)t, &ldt,
                         (const double *)&c_zero, (double *)workvt, &ldworkvt);
                                                 
            // C = C - V W2' = C - V T V' C = (I - V T V') C = H C 
            dgemm( lapack_const(notransV), lapack_const(MagmaTrans),
                         &m, &n, &k,
                         (const double *)&c_neg_one, (double *)v,  &ldv,
                                    (double *)workvt, &ldworkvt,
                         (const double *)&c_one,     (double *)c,       &ldc);
        }
    }
    else {
        // Form C H or C H'
        // Comments assume C H. 
        // When forming C H', T gets transposed via trans.
        
        // W = C V
        dgemm( lapack_const(MagmaNoTrans), lapack_const(notransV),
                     &m, &k, &n,
                     (const double *)&c_one, (double *)c, &ldc,
                             (double *)v, &ldv,
                     (const double *)&c_zero, (double *)work, &ldwork);
        if(m<=n){
            // W2 = W T = C V T
            dgemm( lapack_const(MagmaNoTrans), lapack_const(trans),
                         &m, &k, &k,
                         (const double *)&c_one,  (double *)work, &ldwork,
                                 (double *)t, &ldt,
                         (const double *)&c_zero, (double *)workvt, &ldworkvt);
            
            // C = C - W2 V' = C - C V T V' = C (I - V T V') = C H
            dgemm( lapack_const(MagmaNoTrans), lapack_const(transV),
                         &m, &n, &k,
                         (const double *)&c_neg_one, (double *)work, &ldwork,
                                    (double *)v, &ldv,
                         (const double *)&c_one, (double *)c, &ldc);
        }else{
            // W2 = T V' 
            dgemm( lapack_const(trans), lapack_const(transV),
                         &k, &n, &k,
                         (const double *)&c_one,  (double *)t, &ldt,
                                 (double *)v, &ldv,
                         (const double *)&c_zero, (double *)workvt, &ldworkvt);
            // C = C - W W2 = C - C V T V' = C (I - V T V') = C H
            dgemm( lapack_const(MagmaNoTrans), lapack_const(MagmaNoTrans),
                         &m, &n, &k,
                         (const double *)&c_neg_one, (double *)work, &ldwork,
                                    (double *)workvt, &ldworkvt,
                         (const double *)&c_one, (double *)c, &ldc);

        }
    }
}



void call_dtranspose(magma_mic_dtranspose_param dtranspose_param)
{
        /******************************************************
         * _mm512_load_pd -- for loading
         * _mm512_i32lo[ext]scatter_pd
         * _mm512_i32loscatter_pd is the simplified version of ...ext...
         * they do the same thing, the latter needs less arguments.
         * **************************************************/
        //        printf("entered stranpose\n");
        //        fflush(stdout);
        /*
           double *odata, *idata;
           int ldo,ldi, m, n;
           int num_threads;
           int tid;
           int index[16], i;
           __M512D _reg;
           __m512i _index;
           double *my_head, *loop_head, *dest_head;
           int num_cols, num_cols_g, col, iter, num_iter, num_rows_g, num_rows;

           odata        =        (double *) dtranspose_param.odata;
           ldo                =        dtranspose_param.ldo;
           idata        =        (double *) dtranspose_param.idata;
           ldi                =        dtranspose_param.ldi;
           m                =        dtranspose_param.m;
           n                =        dtranspose_param.n;

           if (m == 0 || n == 0) {
           return;
           }
           if (m <= n) {
           for (i = 0; i < 16; i++) {
           index[i] = i*DREG_WIDTH*ldo;
           }
        // 62 cores, 4 hardware threads per core. Leave out one core for the OS
        // (62 - 1)*4 = 244

        num_threads = 244;
        num_cols_g =        (n + num_threads - 1)/num_threads; 
        num_iter = (int) m*1.0/DREG_WIDTH;
        _index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
        index[11], index[10], index[9], index[8], index[7], index[6],
        index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_cols, iter,\
my_head, loop_head, dest_head, _reg)
{
tid = omp_get_thread_num();
num_cols = num_cols_g;
if (tid == (num_threads - 1) && (n%num_cols_g != 0)) { 
num_cols = n%num_cols_g;
}
// starting point of my section, constant for each thread
my_head = idata + ldi*(tid*num_cols_g);
for (iter = 0; iter < num_iter; iter++) { // loop inside each column
for (col = 0; col < num_cols; col++) { // loop over each column
if (tid*num_cols_g + col < n) {
        // starting point of the inner loop
        loop_head = my_head + col*ldi + iter*DREG_WIDTH;
        dest_head = odata + (iter*DREG_WIDTH)*ldo + (tid*num_cols_g + col);
        // load 8 double precision floating point nubmers from address
        // mt into a float64 vector.
        _reg = _MM512_LOAD_PPD( (void const *)loop_head);
        //                                        _mm512_load_pd( (void const *)loop_head);

        // this intrinsic needs int32 indices, but a 512 register
        // contains 16 int32 integers, and there is only one intrinsic
        // to set all 16 positoins of the register. So _index contains
        // 16 int32 integers.
        _MM512_I32LOEXTSCATTER_PPD((void *) dest_head, _index, _reg,
        _MM_DOWNCONV_PD_NONE, 1, _MM_HINT_NONE);
        }
        }
        }        
        }
        } else {
        // init index[], it's for scattering the entries to the dest matrix
        for (i = 0; i < 16; i++) {
        index[i] = i*DREG_WIDTH*ldo;
        }

        // 62 cores, 4 hardware threads per core. Leave out one core for the OS
        // (62 - 1)*4 = 244
        num_threads = 244;
        // number of rows in each group
        num_rows_g = (m/DREG_WIDTH + num_threads - 1)/num_threads*DREG_WIDTH;
        _index = _mm512_set_epi32(index[15], index[14], index[13], index[12],
                        index[11], index[10], index[9], index[8], index[7], index[6],
                        index[5], index[4], index[3], index[2], index[1], index[0]);

#pragma omp parallel num_threads(num_threads) private(tid, col, num_rows, iter,\
                loop_head, dest_head, _reg, num_iter)
        {
                tid = omp_get_thread_num();
                // compute the number of rows I need to work on
                num_rows = num_rows_g;
                if (tid == (num_threads - 1) && (m%num_rows_g != 0)) {
                        num_rows = m%num_rows_g;
                }
                num_iter = (int) num_rows*1.0/DREG_WIDTH;

                for (iter = 0; iter < num_iter; iter++) {
                        if (tid*num_rows_g + iter*DREG_WIDTH < m) {
                                for (col = 0; col < n; col++) {
                                        loop_head = idata + (tid*num_rows_g + iter*DREG_WIDTH) + ldi*col;
                                        dest_head = odata + ldo*(tid*num_rows_g + iter*DREG_WIDTH) + col;

                                        _reg = _MM512_LOAD_PPD((void const *) loop_head);
                                        _MM512_I32LOEXTSCATTER_PPD((void *) dest_head, _index, _reg,
                                                        _MM_DOWNCONV_PD_NONE, 1, _MM_HINT_NONE);
                                }
                        }
                }

        }
}
*/
return;
}

void call_dtranspose_mkl(magma_mic_dtranspose_param dtranspose_param)
{
    double *odata, *idata;
    int ldo,ldi, m, n;
    int num_threads;
    int tid;
    int index[16], i;
    //  _M512 _reg;
    //  _M512I _index;
    double *my_head, *loop_head, *dest_head;
    int num_cols, num_cols_g, col, iter, num_iter;
    
    odata =(double *) dtranspose_param.odata;
    ldo   =dtranspose_param.ldo;
    idata =(double *) dtranspose_param.idata;
    ldi   = dtranspose_param.ldi;
    m     = dtranspose_param.m;
    n     = dtranspose_param.n;
    double alpha = MAGMA_D_MAKE(  1.0, 0.0 );
    
    mkl_domatcopy('C', 'T', m, n, *((double *)(&alpha)), (double *)idata, ldi, 
                    (double *)odata, ldo);
}

//void call_dlaswp(magma_mic_dlaswp_param dlaswp_param) {
/*
 * takes 0-based ipiv
 * In magma semantics, d_AT is always the starting point of the matrix, it
 * doesn't change.
 */
//        fprintf(stdout, "entering dlaswp\n");
//        double *d_AT;
//        int lda, nb, ind;
//        int *d_ipiv;

//        d_AT                =        (double *) dlaswp_param.d_AT;
//        lda                        =        dlaswp_param.lda;
//        d_ipiv                =        (int *) dlaswp_param.d_ipiv;
//        nb                        =        dlaswp_param.nb;
//        ind                        =        dlaswp_param.ind;

//        if (nb == 0) {
//                return;
//        }

//        int my_i, my_j;
//        int tid;
//        double *p1, *p2, tmp, *head;
//        int  *ipiv_tmp;
//        int num_threads;
//        int blk, num_blk;

//        __M512D _reg1, _reg2;

//        ipiv_tmp = d_ipiv;
// (62 - 1)*4 = 244
//        num_threads                = 244;
// the number of 8-column blocks in each section. Every thread takes one
// section.
//        num_blk                        = (lda/DREG_WIDTH + num_threads - 1)/num_threads;
//        printf("num_blk = %d\n", num_blk);
//#pragma omp parallel num_threads(num_threads) private(tid, p1, p2, tmp, head,\
//                my_i, my_j, blk, _reg1, _reg2)
//        {
/* 
 * num_reps                the number of 512 vectors this thread owns
 */
//                tid = omp_get_thread_num();
//                for (my_i = 0; my_i < nb; my_i++) { // only swap nb rows
// my_j is the destination row index
// convert 1-based to 0 based ipiv
//                        my_j = d_ipiv[my_i + ind] - 1 + ind; 
//                        if (my_i + ind == my_j) continue;
// in my section, work on the 8-column blocks one by one
//                        for (blk = 0; blk < num_blk; blk++) { // loop in a row
//                                head = d_AT + tid*(num_blk*DREG_WIDTH) + blk*DREG_WIDTH;
// check the end of this thread is in range
//                                if (tid*(num_blk*DREG_WIDTH) + blk*DREG_WIDTH < lda) { 
// p1 -- the source element
// p2 -- the dest element
//                                        p1 = head + (my_i + ind)*lda;
//                                        p2 = head + my_j*lda;
//                                        _reg1 = _MM512_LOAD_PPD( (void const *) p1);
//                                        _reg2 = _MM512_LOAD_PPD( (void const *) p2);
//                                        _MM512_STORE_PPD( (void *) p2, _reg1);
//                                        _MM512_STORE_PPD( (void *) p1, _reg2);
//                                }
//                        }
//                }
//        }
//        return;
//}


void call_dlaswp(magma_mic_dlaswp_param dlaswp_param)
{
    magma_int_t n, k1, k2, lda, incx;
    double *d_A;
    magma_int_t *d_ipiv;
                     
    n      = dlaswp_param.n;
    d_A    = (double *) dlaswp_param.d_A;
    lda    = dlaswp_param.lda;
    k1     = dlaswp_param.k1;
    k2     = dlaswp_param.k2;
    d_ipiv = (magma_int_t *) dlaswp_param.d_ipiv;
    incx   = dlaswp_param.incx;
    
    //printf("n: %d lda: %d k1: %d k2: %d incx: %d\n", n, lda, k1, k2, incx);
    dlaswp(&n, (double *)d_A, &lda, &k1, &k2, (magma_int_t *)d_ipiv, &incx);
}


void call_dtrsmgemm(magma_mic_dtrsmgemm_param dtrsmgemm_param)
{
    double gtrsmflops, ggemmflops, mic_trsmtime, mic_gemmtime;

    magma_mic_dtrsm_param trsm_param;
    magma_mic_dgemm_param gemm_param;

    magma_side_t side;
    magma_uplo_t uplo;
    magma_trans_t ttransa;
    magma_diag_t diag;
    magma_int_t tm, tn, tlda, tldb;
    double talpha, *ta, *tb;

    trsm_param = dtrsmgemm_param.trsm_param;

    side = trsm_param.side;
    uplo = trsm_param.uplo;
    ttransa = trsm_param.transa;
    diag = trsm_param.diag;
    tm = trsm_param.m;
    tn = trsm_param.n;
    talpha = trsm_param.alpha;
    ta = (double *)trsm_param.a;
    tlda = trsm_param.lda;
    tb = (double *)trsm_param.b;
    tldb = trsm_param.ldb;

    mic_trsmtime = dsecnd();

    dtrsm( lapack_const(side), lapack_const(uplo), lapack_const(ttransa), lapack_const(diag),
           &tm, &tn, (const double *)&talpha, (double *)ta, &tlda, 
           (double *)tb, &tldb);

    mic_trsmtime = dsecnd() - mic_trsmtime;

    gtrsmflops = 0;
    if(lapack_const(side)[0] == 'L')
        gtrsmflops = MICFLOPS_TRSM_L((double)tm, (double)tn) * 1e-9;
    else if(lapack_const(side)[0] == 'R')
        gtrsmflops = MICFLOPS_TRSM_R((double)tm, (double)tn) * 1e-9;

    magma_trans_t transa, transb;
    magma_int_t m, n, k, lda, ldb, ldc;
    double alpha, beta, *a, *b, *c;

    gemm_param = dtrsmgemm_param.gemm_param;


    transa = gemm_param.transa;
    transb = gemm_param.transb;
    m = gemm_param.m;
    n = gemm_param.n;
    k = gemm_param.k;
    alpha = gemm_param.alpha;
    a = (double *)gemm_param.a;
    lda = gemm_param.lda;
    b = (double *)gemm_param.b;
    ldb = gemm_param.ldb;
    beta = gemm_param.beta;
    c = (double *)gemm_param.c;
    ldc = gemm_param.ldc;

    //printf("DGEMM: %c %c %d %d %d %f 0x%lx %d 0x%lx %d %f 0x%lx %d\n", transa, transb, m, n, k, alpha, (off_t)a, lda, (off_t)b, ldb, beta, (off_t)c, ldc);

    mic_gemmtime = dsecnd();

    dgemm( lapack_const(transa), lapack_const(transb), 
           &m, &n, &k, (const double *)&alpha, (double *)a, &lda, 
           (double *)b, &ldb, (const double *)&beta, 
           (double *)c, &ldc);

    mic_gemmtime = dsecnd() - mic_gemmtime;

    ggemmflops = MICFLOPS_GEMM((double)m, (double)n, (double)k) * 1e-9;

    //printf("DTRSMGEMM: %8.4lf( %8.4lf )\t %8.4lf( %8.4lf )\n", gtrsmflops / mic_trsmtime, mic_trsmtime * 1000, ggemmflops / mic_gemmtime, mic_gemmtime * 1000); 
}
