/*
 *  -- MAGMA (version 1.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     April 2011
 *
 * @generated c Thu May  2 22:50:45 2013
 *
 **/
// includes, system
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "magma_types.h"
#include "testings.h"

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing cgesv_mic
*/
int main(int argc , char **argv)
{
    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    magma_int_t num = 0;
    magma_err_t err;

    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
	    fprintf( stderr, "magma_get_devices failed: %d\n", err );
	    exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
	    fprintf( stderr, "magma_queue_create failed: %d\n", err );
	    exit(-1);
    }

    real_Double_t   gflops, mic_perf, mic_time;
    float          Rnorm, Anorm, Xnorm, *work;
    magmaFloatComplex c_one     = MAGMA_C_ONE;
    magmaFloatComplex c_neg_one = MAGMA_C_NEG_ONE;
    magmaFloatComplex *h_A, *h_B, *h_X;
    magmaFloatComplex *d_A, *d_B;
    magma_int_t *ipiv;
    magma_int_t N, NRHS, lda, ldb, ldda, lddb, info, sizeA, sizeB;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    
    magma_opts opts;
    parse_opts( argc, argv, &opts );
    
    NRHS = opts.nrhs;
    
    printf("    N   NRHS   MAGMA GFlop/s (sec)   ||B - AX|| / ||A||*||X||\n");
    printf("===========================================================\n");
    for( int i = 0; i < opts.ntest; ++i ) {
        for( int iter = 0; iter < opts.niter; ++iter ) {
            N = opts.nsize[i];
            lda    = N;
            ldb    = lda;
            ldda   = ((N+31)/32)*32;
            lddb   = ldda;
            gflops = ( FLOPS_CGETRF( N, N ) + FLOPS_CGETRS( N, NRHS ) ) / 1e9;
            
            TESTING_MALLOC_HOST( h_A, magmaFloatComplex, lda*N    );
            TESTING_MALLOC_HOST( h_B, magmaFloatComplex, ldb*NRHS );
            TESTING_MALLOC_HOST( h_X, magmaFloatComplex, ldb*NRHS );
            TESTING_MALLOC_HOST( work, float,         N        );
            TESTING_MALLOC_HOST( ipiv, magma_int_t,    N        );
            
            TESTING_MALLOC_DEV( d_A, magmaFloatComplex, ldda*N    );
            TESTING_MALLOC_DEV( d_B, magmaFloatComplex, lddb*NRHS );
            
            /* Initialize the matrices */
            sizeA = lda*N;
            sizeB = ldb*NRHS;
            lapackf77_clarnv( &ione, ISEED, &sizeA, h_A );
            lapackf77_clarnv( &ione, ISEED, &sizeB, h_B );
            
            magma_csetmatrix( N, N,    h_A, 0, lda, d_A, 0, ldda, queue );
            magma_csetmatrix( N, NRHS, h_B, 0, ldb, d_B, 0, lddb, queue );
            
            //=====================================================================
            // Solve Ax = b through an LU factorization, using MAGMA
            //=====================================================================
            mic_time = magma_wtime();
            magma_cgesv_mic( N, NRHS, d_A, 0, ldda, ipiv, d_B, 0, lddb, &info, queue );
            mic_time = magma_wtime() - mic_time;
            mic_perf = gflops / mic_time;
            if (info != 0)
                printf("magma_cgesv_mic returned error %d.\n", (int) info);
            
            //=====================================================================
            // Residual
            //=====================================================================
            magma_cgetmatrix( N, NRHS, d_B, 0, lddb, h_X, 0, ldb, queue );
            
            Anorm = lapackf77_clange("I", &N, &N,    h_A, &lda, work);
            Xnorm = lapackf77_clange("I", &N, &NRHS, h_X, &ldb, work);
            
            blasf77_cgemm( MagmaNoTransStr, MagmaNoTransStr, &N, &NRHS, &N,
                           &c_one,     h_A, &lda,
                                       h_X, &ldb,
                           &c_neg_one, h_B, &ldb);
            
            Rnorm = lapackf77_clange("I", &N, &NRHS, h_B, &ldb, work);
            
            printf( "%5d  %5d   %7.2f (%7.2f)   %8.2e\n",
                    (int) N, (int) NRHS, mic_perf, mic_time, Rnorm/(Anorm*Xnorm) );
            
            TESTING_FREE_HOST( h_A );
            TESTING_FREE_HOST( h_B );
            TESTING_FREE_HOST( h_X );
            TESTING_FREE_HOST( work );
            TESTING_FREE_HOST( ipiv );
            
            TESTING_FREE_DEV( d_A );
            TESTING_FREE_DEV( d_B );
        }
        if ( opts.niter > 1 ) {
            printf( "\n" );
        }
    }

    /* Shutdown */

    magma_queue_destroy( queue );
    magma_finalize();
    return 0;
}
