/*
 *  -- micMAGMA (version 1.3.0) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     April 2012
 *
 *     @generated ds Thu May  2 22:50:45 2013
 *
 **/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"

#define PRECISION_d

int main(int argc , char **argv)
{
    real_Double_t time;
    double      gflopsF, gflopsS, gpu_perf;
    double      gpu_perfdf, gpu_perfds;
    double      gpu_perfsf, gpu_perfss;
    double      Rnorm, Anorm;
    double c_one     = MAGMA_D_ONE;
    double c_neg_one = MAGMA_D_NEG_ONE;
    double *h_A, *h_A2, *h_B, *h_X;
    magmaDouble_ptr d_A, d_B, d_X;
    double *WORKD;

    magmaFloat_ptr  d_As, d_Bs;
    float *WORKS;

    double          *h_workd;
    magma_int_t *h_ipiv, *d_ipiv;
    magma_int_t lda, ldb, ldx;
    magma_int_t ldda, lddb, lddx;
    magma_int_t i, iter, info, size;
    magma_int_t N        = 0;
    magma_int_t ione     = 1;
    magma_int_t NRHS     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    magma_int_t sizetest[10] = {1024,2048,3072,4032,5184,6016,7040,7520,8064,8192};

    char trans = MagmaNoTrans;
    //char trans = MagmaTrans;

    if (argc != 1){
        for(i = 1; i<argc; i++){        
            if (strcmp("-N", argv[i])==0)
                N = atoi(argv[++i]);
            else if (strcmp("-nrhs", argv[i])==0)
                NRHS = atoi(argv[++i]);
        }
        if (N>0) sizetest[0] = sizetest[9] = N;
        else exit(1);
    }
    else {
        printf("\nUsage: \n");
        printf("  testing_dsgesv_mic -nrhs %d -N %d\n\n", (int) NRHS, 1024);
    }

    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;

    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", err );
      exit(-1);
    }

    
    printf("Epsilon(double): %8.6e\n"
           "Epsilon(single): %8.6e\n\n", 
           lapackf77_dlamch("Epsilon"), lapackf77_slamch("Epsilon") );

    N = sizetest[9];
    ldb  = ldx = lda = N;
    ldda = ((N+31)/32)*32;
    lddb = lddx = ldda;

    TESTING_MALLOC_HOST(  h_A, double, lda*N    );
    TESTING_MALLOC_HOST( h_A2, double, lda*N    );

    TESTING_MALLOC_HOST( h_B, double, ldb*NRHS );
    TESTING_MALLOC_HOST( h_X, double, ldx*NRHS );
    TESTING_MALLOC_HOST( h_ipiv, magma_int_t,    N        );
    TESTING_MALLOC_HOST( h_workd, double, N );
    
    TESTING_MALLOC_DEV( d_A,     double, ldda*N        );
    TESTING_MALLOC_DEV( d_B,     double, lddb*NRHS     );

    TESTING_MALLOC_DEV( d_As,    float, ldda*N        );
    TESTING_MALLOC_DEV( d_Bs,    float, lddb*NRHS     );

    TESTING_MALLOC_DEV( d_X,     double, lddx*NRHS     );
    TESTING_MALLOC_DEV( d_ipiv,  magma_int_t,     N             );

    TESTING_MALLOC_HOST( WORKS, float,  ldda*(N+NRHS) );
    TESTING_MALLOC_HOST( WORKD, double, N*NRHS        );

    printf("  N   DP-Factor  DP-Solve  SP-Factor  SP-Solve  MP-Solve  ||b-Ax||/||A||  NumIter\n");
    printf("==================================================================================\n");
    for(i=0; i<10; i++){
        N = sizetest[i] ;
        
        gflopsF = FLOPS_ZGETRF( N, N ) / 1e9;
        gflopsS = gflopsF + ( FLOPS_ZGETRS( N, NRHS ) / 1e9 );

        ldb  = ldx = lda = N;
        ldda = ((N+31)/32)*32;
        lddb = lddx = N;//ldda;

        /* Initialize matrices */
        size = lda * N;
        lapackf77_dlarnv( &ione, ISEED, &size, h_A );
        size = ldb * NRHS;
        lapackf77_dlarnv( &ione, ISEED, &size, h_B );

        lapackf77_dlacpy( MagmaUpperLowerStr, &N, &NRHS, h_B, &ldb, h_X, &ldx);
	lapackf77_dlacpy( MagmaUpperLowerStr, &lda, &N, h_A, &lda, h_A2, &lda);

        printf("%5d  ",(int) N);

        magma_dsetmatrix( N, N,    h_A, 0, lda, d_A, 0, ldda, queue );
        magma_dsetmatrix( N, NRHS, h_B, 0, ldb, d_B, 0, lddb, queue );

        //=====================================================================
        //              MIXED - GPU
        //=====================================================================
	magma_kernel_sync();
	lapackf77_dlag2s( &N, &N, h_A2, &lda, WORKS, &N, &info );
	magma_sgetrf2(N, N, WORKS, N, h_ipiv, d_As, ldda, d_ipiv, &info, queue);
	magma_kernel_sync();
	
        time = magma_wtime();
        magma_dsgesv2(trans, N, NRHS, 
		      h_A2, lda, 
		      d_As, 0, ldda, h_ipiv, d_ipiv, 
		      h_B, ldb, h_X, ldx, 
		      WORKD, WORKS, &iter, &info, queue );
	magma_kernel_sync();
        time = magma_wtime() - time;
        if (info < 0)
            printf("Argument %d of magma_dsgesv had an illegal value.\n", (int) -info);
        gpu_perf = gflopsS / time;

        //=====================================================================
        //              ERROR DP vs MIXED  - GPU
        //=====================================================================
        Anorm = lapackf77_dlange("I", &N, &N, h_A, &lda, h_workd);
        blasf77_dgemm( lapack_const(trans), MagmaNoTransStr, 
                       &N, &NRHS, &N, 
                       &c_one,     h_A, &lda,
                                   h_X, &ldx,
                       &c_neg_one, h_B, &ldb);
        Rnorm = lapackf77_dlange("I", &N, &NRHS, h_B, &ldb, h_workd);

        //=====================================================================
        //                 Double Precision Factor 
        //=====================================================================
        magma_dsetmatrix( N, N, h_A, 0, lda, d_A, 0, ldda, queue );
        
	magma_kernel_sync();
        time = magma_wtime();
        magma_dgetrf_mic(N, N, d_A, 0, ldda, h_ipiv, &info, queue );
	magma_kernel_sync();
        time = magma_wtime() - time;
        if (info < 0)
            printf("Argument %d of magma_dgetrf had an illegal value.\n", (int) -info);
        gpu_perfdf = gflopsF / time;

        printf("%6.2f    ", gpu_perfdf); fflush(stdout);

        //=====================================================================
        //                 Double Precision Solve 
        //=====================================================================
        magma_dsetmatrix( N, N,    h_A, 0, lda, d_A, 0, ldda, queue );
        magma_dsetmatrix( N, NRHS, h_B, 0, ldb, d_B, 0, lddb, queue );

	magma_kernel_sync();
        time = magma_wtime();
        magma_dgetrf_mic(N, N, d_A, 0, ldda, h_ipiv, &info, queue );
        magma_dgetrs_mic( trans, N, NRHS, d_A, 0, ldda, h_ipiv, d_B, 0, lddb, &info, queue );
	magma_kernel_sync();
	magma_kernel_sync();
        time = magma_wtime() - time;
        if (info < 0)
            printf("Argument %d of magma_dgetrs had an illegal value.\n", (int) -info);

        gpu_perfds = gflopsS / time;

        printf("%6.2f    ", gpu_perfds); fflush(stdout);

        //=====================================================================
        //                 Single Precision Factor 
        //=====================================================================
        magma_dsetmatrix( N, N,    h_A, 0, lda,  d_A, 0, ldda, queue );
        magma_dsetmatrix( N, NRHS, h_B, 0, ldb,  d_B, 0, lddb, queue );

        magma_dlag2s( N, N,    d_A, 0, ldda, d_As, 0, ldda, queue ); 
        magma_dlag2s( N, NRHS, d_B, 0, lddb, d_Bs, 0, lddb, queue );

	magma_kernel_sync();
        time = magma_wtime();
        magma_sgetrf_mic(N, N, d_As, 0, ldda, h_ipiv, &info, queue );
	magma_kernel_sync();
        time = magma_wtime() - time;
        if (info < 0)
            printf("Argument %d of magma_sgetrf had an illegal value.\n", (int) -info);
        
        gpu_perfsf = gflopsF / time;
        printf("%6.2f     ", gpu_perfsf); fflush(stdout);

        //=====================================================================
        //                 Single Precision Solve 
        //=====================================================================
        magma_dlag2s(N, N,    d_A, 0, ldda, d_As, 0, ldda, queue ); 
        magma_dlag2s(N, NRHS, d_B, 0, lddb, d_Bs, 0, lddb, queue );

	magma_kernel_sync();
        time = magma_wtime();
        magma_sgetrf_mic( N, N,    d_As, 0, ldda, h_ipiv, &info, queue );
        magma_sgetrs_mic( trans, N, NRHS, d_As, 0, ldda, h_ipiv, 
                          d_Bs, 0, lddb, &info, queue );
	magma_kernel_sync();
        time = magma_wtime() - time;
        if (info < 0)
            printf("Argument %d of magma_sgetrs had an illegal value.\n", (int) -info);

        gpu_perfss = gflopsS / time;
        printf("%6.2f     ", gpu_perfss); fflush(stdout);

        printf("%6.2f     ", gpu_perf);
        printf("%e    %3d\n", Rnorm/Anorm, (int) iter); fflush(stdout);

        if( argc != 1 ) break ;        
    }

    /* Memory clean up */
    TESTING_FREE_HOST( h_A );
    TESTING_FREE_HOST( h_A2 );

    TESTING_FREE_HOST( h_B );
    TESTING_FREE_HOST( h_X );
    TESTING_FREE_HOST( h_ipiv );
    TESTING_FREE_HOST( h_workd );
    
    TESTING_FREE_DEV( d_A );
    TESTING_FREE_DEV( d_B );

    TESTING_FREE_DEV( d_As );
    TESTING_FREE_DEV( d_Bs );

    TESTING_FREE_DEV( d_X );
    TESTING_FREE_DEV( d_ipiv );

    TESTING_FREE_HOST( WORKS );
    TESTING_FREE_HOST( WORKD );

    /* Shutdown */
    magma_queue_destroy( queue );
    magma_finalize();
}
