/*
 *  -- micMAGMA (version 0.1) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     April 2012
 *
 * @generated s Thu May  2 22:50:42 2013
 *
 **/

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"


/* ////////////////////////////////////////////////////////////////////////////
   -- Testing ssyr2k
*/
int main( int argc, char** argv)
{
    real_Double_t   gflops, mic_perf, mic_time, cpu_perf, cpu_time;
    float          mic_error, Cnorm, work[1];
    magma_int_t N, K;
    magma_int_t Ak, An, Bk, Bn;
    magma_int_t sizeA, sizeB, sizeC;
    magma_int_t lda, ldb, ldc, ldda, lddb, lddc;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    
    float *h_A, *h_B, *h_C, *h_Cmic;
    float *d_A, *d_B, *d_C;
    float c_neg_one = MAGMA_S_NEG_ONE;
    float alpha = MAGMA_S_MAKE(  0.29, -0.86 );
    float beta  = MAGMA_D_MAKE( -0.48,  0.38 );
    
    magma_opts opts;
    parse_opts( argc, argv, &opts );

    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;
    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", err );
      exit(-1);
    } 
   
    printf("    N     K   MIC Gflop/s (ms)   CPU Gflop/s (ms)  MIC error\n");
    printf("==================================================================\n");
 
   for( int i = 0; i < opts.ntest; ++i ) {
        for( int iter = 0; iter < opts.niter; ++iter ) {
            N = opts.nsize[i];
            K = opts.ksize[i];
            gflops = FLOPS_SSYR2K(K, N) / 1e9;

            if ( opts.transA == MagmaNoTrans ) {
                lda = An = N;
                Ak = K;
                ldb = Bn = N;
                Bk = K;
            } else {
                lda = An = K;
                Ak = N;
                ldb = Bn = K;
                Bk = N;
            }
            
            ldc = N;
            
            ldda = ((lda+31)/32)*32;
            lddb = ((ldb+31)/32)*32;
            lddc = ((ldc+31)/32)*32;
            
            sizeA = lda*Ak;
            sizeB = ldb*Ak;
            sizeC = ldc*N;
            
            TESTING_MALLOC_HOST( h_A,  float, lda*Ak );
            TESTING_MALLOC_HOST( h_B,  float, ldb*Bk );
            TESTING_MALLOC_HOST( h_C,  float, ldc*N  );
            TESTING_MALLOC_HOST( h_Cmic, float, ldc*N  );
            
            TESTING_MALLOC_DEV( d_A, float, ldda*Ak );
            TESTING_MALLOC_DEV( d_B, float, lddb*Bk );
            TESTING_MALLOC_DEV( d_C, float, lddc*N  );
            
            /* Initialize the matrices */
            lapackf77_slarnv( &ione, ISEED, &sizeA, h_A );
            lapackf77_slarnv( &ione, ISEED, &sizeB, h_B );
            lapackf77_slarnv( &ione, ISEED, &sizeC, h_C );
            
            /* =====================================================================
               Performs operation using MIC-BLAS
               =================================================================== */
            magma_ssetmatrix( An, Ak, h_A, 0, lda, d_A, 0, ldda, queue );
            magma_ssetmatrix( Bn, Bk, h_B, 0, ldb, d_B, 0, lddb, queue );
            magma_ssetmatrix( N, N, h_C, 0, ldc, d_C, 0, lddc, queue );
            
            mic_time = magma_sync_wtime( NULL );
            magma_ssyr2k( opts.uplo, opts.transA, N, K,
			  alpha, d_A, 0, ldda,
			         d_B, 0, lddb,
                          beta,  d_C, 0, lddc, queue );
            mic_time = magma_sync_wtime( NULL ) - mic_time;
            mic_perf = gflops / mic_time;
            
	    /* =====================================================================
	       Performs operation using CPU BLAS
	       =================================================================== */
	    if ( opts.check ) {
		    magma_sgetmatrix( N, N, d_C, 0, lddc, h_Cmic, 0, ldc, queue );

		    cpu_time = magma_wtime();
		    blasf77_ssyr2k( lapack_const(opts.uplo), lapack_const(opts.transA), &N, &K,
				    &alpha, h_A, &lda,
				    h_B, &ldb,
				    &beta,  h_C, &ldc );
		    cpu_time = magma_wtime() - cpu_time;
		    cpu_perf = gflops / cpu_time;

		    /* =====================================================================
		       Check the result
		       =================================================================== */
		    // compute relative error for both magma & mic, relative to lapack,
		    // |C_magma - C_lapack| / |C_lapack|
		    Cnorm = lapackf77_slange( "M", &N, &N, h_C, &ldc, work );

		    blasf77_saxpy( &sizeC, &c_neg_one, h_C, &ione, h_Cmic, &ione );
		    mic_error = lapackf77_slange( "M", &N, &N, h_Cmic, &ldc, work ) / Cnorm;

		    printf("%5d %5d   %7.2f (%7.2f)   %7.2f (%7.2f)    %8.2e\n",
				    (int) N, (int) K,
				    mic_perf, 1000.*mic_time,
				    cpu_perf,    1000.*cpu_time,
				    mic_error );
	    }
	    else {
		    printf("%5d %5d   %7.2f (%7.2f)    ---   (  ---  )    ---\n",
				    (int) N, (int) K,
				    mic_perf, 1000.*mic_time);
	    }

	    TESTING_FREE_HOST( h_A  );
	    TESTING_FREE_HOST( h_B  );
	    TESTING_FREE_HOST( h_C  );
	    TESTING_FREE_HOST( h_Cmic );

	    TESTING_FREE_DEV( d_A );
	    TESTING_FREE_DEV( d_B );
	    TESTING_FREE_DEV( d_C );
	}
	if ( opts.niter > 1 ) {
		printf( "\n" );
	}
   }

   magma_queue_destroy( queue );
   magma_finalize();
   return 0;
}
